from __future__ import unicode_literals

import six
import abc
import json
import logging

from jsonpointer import resolve_pointer, JsonPointerException

logger = logging.getLogger(__name__)

try:
    from lxml import etree
except ImportError:
    logger.info('You will need to install lxml to use the XML transformer')


@six.add_metaclass(abc.ABCMeta)
class BaseTransformer(object):

    def __init__(self, schema):
        self.schema = schema

    @abc.abstractmethod
    def _transform_string(self, string, doc):
        raise NotImplementedError

    @abc.abstractmethod
    def load(self, doc):
        raise NotImplementedError

    def transform(self, doc, fail=False, load=True):
        doc = self.load(doc) if load else doc
        return self._transform_dict(self.schema, doc, fail=fail)

    def _transform_dict(self, d, doc, fail=False):
        return {
            key: self._maybe_transform_value(value, doc, fail=fail)
            for key, value in d.items()
        }

    def _transform_list(self, l, doc, fail=False):
        return [
            self._maybe_transform_value(item, doc, fail=fail)
            for item in l
        ]

    def _maybe_transform_value(self, value, doc, fail=False):
        try:
            return self._transform_value(value, doc, fail=fail)
        except Exception as e:
            if fail:
                raise
            logger.exception(e)
            return None

    def _transform_value(self, value, doc, fail=False):
        if isinstance(value, dict):
            return self._transform_dict(value, doc, fail=fail)
        elif isinstance(value, list):
            return self._transform_list(value, doc, fail=fail)
        elif isinstance(value, tuple) and len(value) == 2 and isinstance(value[0], tuple):
            return self._transform_args_kwargs(value, doc)
        elif isinstance(value, tuple):
            return self._transform_tuple(value, doc)
        elif isinstance(value, six.string_types):
            return self._transform_string(value, doc)
        elif callable(value):
            return value(doc)

    def _transform_tuple(self, l, doc):

        fn, values = l[-1], l[:-1]
        args = []

        for value in values:
            if isinstance(value, six.string_types):
                args.append(self._transform_string(value, doc))
            elif callable(value):
                args.append(value(doc))
        return fn(*args)

    def _transform_args_kwargs(self, l, doc):
        fn = l[1]
        return fn(
            *self._transform_args(l[0], doc),
            **self._transform_kwargs(l[0], doc)
        )

    def _transform_args(self, t, doc):
        return [self._transform_string(arg, doc) for arg in t[0]]

    def _transform_kwargs(self, t, doc):
        return {
            k: self._transform_string(v, doc) for k, v in t[1].items()
        } if len(t) == 2 else {}


@six.add_metaclass(abc.ABCMeta)
class XMLTransformer(BaseTransformer):

    def __init__(self, schema, namespaces=None):
        BaseTransformer.__init__(self, schema)
        self.namespaces = namespaces or {}

    def load(self, doc):
        return etree.XML(doc)

    def _transform_string(self, string, doc):
        return doc.xpath(string, namespaces=self.namespaces)


@six.add_metaclass(abc.ABCMeta)
class JSONTransformer(BaseTransformer):

    def load(self, doc):
        return json.loads(doc)

    def _transform_string(self, val, doc):
        try:
            return resolve_pointer(doc, val)
        except JsonPointerException as e:
            # This is because of jsonpointer's exception structure
            if 'not found in' in e.args[0] or 'is not a valid list index' in e.args[0]:
                return None
            raise e


@six.add_metaclass(abc.ABCMeta)
class CSVTransformer(BaseTransformer):

    def __init__(self, schema, keys):
        BaseTransformer.__init__(self, schema)
        self.keys = dict((val, key) for key, val in enumerate(keys))

    def load(self, doc):
        raise NotImplementedError

    def _transform_string(self, val, doc):
        return doc[self.keys[val]]



import functools
from copy import deepcopy


def CONSTANT(x):
    ''' Takes a value, returns a function that always returns that value
        Useful inside schemas for defining constants

        >>> CONSTANT(7)('my', 'name', verb='is')
        7
        >>> CONSTANT([123, 456])()
        [123, 456]
    '''
    def inner(*y, **z):
        return x
    return inner


def compose(*functions):
    ''' evaluates functions from right to left.

        >>> add = lambda x, y: x + y
        >>> add3 = lambda x: x + 3
        >>> divide2 = lambda x: x/2
        >>> subtract4 = lambda x: x - 4
        >>> subtract1 = compose(add3, subtract4)
        >>> subtract1(1)
        0
        >>> compose(subtract1, add3)(4)
        6
        >>> compose(int, add3, add3, divide2)(4)
        8
        >>> compose(int, divide2, add3, add3)(4)
        5
        >>> compose(int, divide2, compose(add3, add3), add)(7, 3)
        8
    '''
    def inner(func1, func2):
        return lambda *x, **y: func1(func2(*x, **y))
    return functools.reduce(inner, functions)


def updated_schema(old, new):
    ''' Creates a dictionary resulting from adding all keys/values of the second to the first
        The second dictionary will overwrite the first.

        >>> old, new = {'name': 'ric', 'job': None}, {'name': 'Rick'}
        >>> updated = updated_schema(old, new)
        >>> len(updated.keys())
        2
        >>> print(updated['name'])
        Rick
        >>> updated['job'] is None
        True

    '''
    d = deepcopy(old)
    for key, value in new.items():
        if isinstance(value, dict) and old.get(key) and isinstance(old[key], dict):
            d[key] = updated_schema(old[key], new[key])
        else:
            d[key] = value
    return d


def single_result(l, default=''):
    ''' A function that will return the first element of a list if it exists

        >>> print(single_result(['hello', None]))
        hello
        >>> print(single_result([], default='hello'))
        hello
        >>> print(single_result([]))
        <BLANKLINE>

    '''
    return l[0] if l else default



from setuptools import setup

VERSION = '0.0.2'

install_reqs = [
    'jsonpointer',
    'jsonschema',
    'six'
]

setup(
    name='schema-transformer',
    packages=['schema_transformer'],
    version=VERSION,
    description='A library for doing schema transformations',
    author='Fabian von Feilitzsch',
    author_email='fabian@cos.io',
    url='https://github.com/fabianvf/schema-transformer',
    download_url='https://github.com/fabianvf/schema-transformer/tarball/{}'.format(VERSION),
    install_requires=install_reqs
)
